import os
import pyinotify

from synspark_logger import api
from synspark_logger.output import log

# module vars
__monitor = None
__notifier = None
__watcher = {}


# public functions
def init():
    """Init log file watcher."""
    global __monitor
    __monitor = pyinotify.WatchManager()
    log.debug('Pyinotify initialized.')


def start():
    """Start log file watcher."""
    global __notifier
    __notifier = pyinotify.ThreadedNotifier(__monitor, ProcessPyinotify())
    __notifier.start()
    log.debug('Pyinotify started.')


def stop():
    """Stop log file watcher."""
    global __notifier, __monitor
    __notifier.stop()  # Stop the notifier thread
    __notifier.join()  # wait done
    __notifier = None
    __monitor = None
    log.debug('Pyinotify stopped.')


def watcher_add(path, read_all=False, filter={}):
    """Add path from watcher.

    :param path:  log path
    :param read_all: on file created: start read to begin
    :param filter: filters rules object
    """
    # clean value
    path = path.strip()
    if not path or path in __watcher:
        return

    # parent directory (log rotate,, create, ...)
    if os.path.isdir(path):
        mask = pyinotify.IN_CREATE | pyinotify.IN_MOVED_TO
        __watcher[path] = {
            'watch': __monitor.add_watch(path, mask),
        }

    # file
    else:
        watcher_add(os.path.dirname(path))  # add parent directory

        if os.path.isfile(path):
            seek = 0 if read_all else os.stat(path).st_size
            mask = pyinotify.IN_MODIFY | pyinotify.IN_DELETE_SELF
            watch = __monitor.add_watch(path, mask).values()[0]

        else:  # file not exist for the moment
            seek = 0
            watch = None

        __watcher[path] = {
            'seek': seek,
            'watch': watch,
            'filter': filter,
        }
    log.debug('Pyinotify: Watcher add %s' % path)


def watcher_del(path):
    """Remove path from watcher.

    :param path:  log path
    :return: is cleaned remove
    """
    if path not in __watcher:
        return False

    result = True

    if __watcher[path]['watch']:
        result &= __monitor.rm_watch(__watcher[path]['watch']).values()[0]

    del(__watcher[path])
    log.debug('Pyinotify: Watcher del %s', path)
    return result


# private functions
class ProcessPyinotify(pyinotify.ProcessEvent):
    """Pyinotify event handler."""
    def process_default(self, event):
        try:
            _callback(event)
        except Exception as e:
            log.error('Error in ProcessPyinotify: %s' % e, exc_info=True)


def _callback(event):
    """Calback for process Pyinotify event.

    :param event: Pyinotify event
    """
    log.debug('Pyinotify: callback for %s', event)
    path = event.pathname

    if event.mask & pyinotify.IN_ISDIR:
        log.debug('Pyinotify: Ignoring directory event at %s' % path)
        return

    # create or moved: log rotate, new process, ...
    if event.mask & (pyinotify.IN_CREATE | pyinotify.IN_MOVED_TO |
                     pyinotify.IN_DELETE_SELF):
        if path not in __watcher:
            log.debug('Pyinotify: Ignoring of %s we do not monitor' % path)
            return

        # refresh watcher
        watcher_del(path)
        watcher_add(path, read_all=True)  # file created: start read to begin

    # read new log lines
    _read_new_line(path)


def _read_new_line(path):
    """Read new log file and send to api.

    :param path: log path
    """
    if not os.path.isfile(path):  # don't read on deleted events
        return

    with open(path, 'r') as f:
        f.seek(__watcher[path]['seek'])  # set position at last read
        filter = __watcher[path]['filter']

        # get new lines
        for line in f.readlines():
            line = line.strip()

            if filter.check(message=line):
                api.send(log=line, path=path)

    # save new position
    __watcher[path]['seek'] = os.stat(path).st_size



import re

from synspark_logger.output import log


class Filter:
    # builting
    def __init__(self, name='NO_NAME'):
        """Init filter

        :param name: name of filter (for identifiate / debug)
        """
        self.name = name
        self.__message_rules = {}  # message's filter
        self.__min_criticity = -1  # to allow (disable by default)
        self.__syslog_facility = []

    def __len__(self):
        """Get number of rules

        :return: rules len
        """
        return len(self.__message_rules) + len(self.__syslog_facility) + \
               (1 if self.__min_criticity >= 0 else 0)  # min criticity are set ?

    def __repr__(self):
        """Get object representation"""
        return '<%s(%s) size=%d>' % (self.__class__.__name__, self.name, len(self))

    # public method
    def add_message_rule(self, rule):
        """ Add a filter string

        :param rule: string rule (for message check)
        """
        filter_type, filter_rule = rule.split(':', 1)

        if filter_type not in self.__message_rules:
            self.__message_rules[filter_type] = []

        # compile regex filters
        if 'regex' in filter_type:
            filter_rule = re.compile(filter_rule)

        self.__message_rules[filter_type].append(filter_rule)

    def add_syslog_facility(self, *args):
        """Add syslog facility (for syslog facility check)

        :param args: list of syslog facility (id)
        """
        self.__syslog_facility += args

    def check(self, message=None, criticity=None, syslog_facility=None):
        """Check if it's allowned to sent to api

        :param message: value string to check
        :param criticity: criticity name
        :return: is allowned
        """
        return self.is_empty() or \
               (message is None or self.__check_message(message)) and \
               (criticity is None or self.__check_criticity(criticity)) and \
               (syslog_facility is None or self.__check_syslog_facility(syslog_facility))

    def is_empty(self):
        """This filter is empty ? (no rule)

        :return: is empty boolean
        """
        return len(self) == 0

    # private method

    def set_min_criticity(self, criticity):
        """ set the minimum level to send to api

        :param criticity: criticity name
        """
        self.__min_criticity = self.__get_criticity_value(criticity)

    def __check_criticity(self, criticity):
        """Check if criticity is allowned to sent to api

        :param criticity: criticity name
        :return: is allowned
        """
        return self.__min_criticity >= self.__get_criticity_value(criticity)

    def __check_message(self, message):
        """Check if value is allowned to sent to api (apply filter rules)

        :param message: value string to check
        :return: is allowned
        """
        if not self.__message_rules:  # no filter
            return False

        for filter_type in self.__message_rules:
            check_handler = {
                'contain':      lambda filter: filter in message,
                'end_with':     lambda filter: message.endswith(filter),
                'regex':        lambda filter: filter.search(message),
                'start_with':   lambda filter: message.startswith(filter),
            }.get(filter_type)

            # unknown filter type
            if not check_handler:
                log.error('%s: %s unknown' % (self, filter_type))
                continue

            # filter check
            for filter in self.__message_rules[filter_type]:
                if check_handler(filter):
                    return True

    def __check_syslog_facility(self, syslog_facility):
        return syslog_facility in self.__syslog_facility

    def __get_criticity_value(self, criticity):
        """Get (syslog) criticity code

        :param criticity: criticity name
        :return: criticity int code
        """
        criticity = {
            'EMERGENCY':    0,
            'ALERT':        1,
            'CRITICAL':     2,
            'ERROR':        3,
            'WARNING':      4,
            'NOTICE':       5,
        }.get(str(criticity).upper(), criticity)

        if isinstance(criticity, int):
            return criticity

        raise TypeError('%s: criticity "%s" in unknown' % (self, criticity))



import logging
import sys


# Gets the instance of the logger.
log = logging.getLogger('Synspark_logger')


# Init output log: redirect log to cli and edit format
if not(len(log.handlers)):
    logout = logging.StreamHandler(sys.stdout)
    logout.setFormatter(logging.Formatter('%(levelname)-6s %(message)s'))
    log.addHandler(logout)


import hashlib
import json
import os
import threading
import requests
import time
from .output import log


key = ''
url = 'https://api.synspark.com/v1.0/events/log'
max_retry = 10
retry_dir_path = ''


def start():
    if not retry_dir_path:
        log.critical('Retry dir path not found!')
        raise ValueError('Retry dir path not found!')

    if not os.path.exists(retry_dir_path):
        os.makedirs(retry_dir_path)

    sleep = 0
    count = 0
    for key in os.listdir(retry_dir_path):
        run_retry(key, sleep)
        count += 1
        sleep = int(count / 10)


def register_retry(data):
    global _retry

    h = hashlib.md5()
    h.update(str(data))
    key = '%s_%s' % (time.time(), h.hexdigest())
    path = os.path.join(retry_dir_path, key)

    with open(path, 'w+') as f:
        json.dump({'data': data, 'retry': 1}, f)

    run_retry(key)


def run_retry(key, sleep=0):
    threading.Thread(None, _run_retry, 'retry', (key, sleep)).start()


def send(_is_retry=False, **data):
    try:
        result = requests.post(url, data, headers={'token': key})
    except Exception as e:
        try:
            log.error('On API: %s' %  e)
            if not _is_retry:
                register_retry(data)
            return False
        except:
            raise e

    log.debug('%s %s %s' % (result.status_code, result.content, data))

    # unrecoverable error
    if 400 <= result.status_code < 500 or result.status_code == 501:
        log.error('%s statuts code(%s), data: %s' % (
            result.status_code, result.content, data
        ))
        return False

    elif result.status_code >= 500 and not _is_retry:
        log.error('%s add rety(%s), data: %s' % (
            result.status_code, result.content, data
        ))
        register_retry(data)
        return False

    return True


def _run_retry(key, sleep):
    if sleep > 0:
        time.sleep(sleep)

    path = os.path.join(retry_dir_path, key)

    with open(path, 'r') as f:
        infos = json.load(f)

    while True:
        if infos['retry'] > max_retry:
            os.remove(path)
            log.error('API max retry for: %s' % infos['data'])
            return True

        log.debug('retry for: %s' % infos['data'])
        if send(_is_retry=True, **infos['data']):
            os.remove(path)
            return True

        infos['retry'] += 1
        with open(path, 'w+') as f:
            json.dump(infos, f)
        time.sleep(infos['retry'] * 60)



from . import api
from .backend import ALL as ALL_BACKEND
from .output import log


__version__ = '1.1.10'
NAME = 'Synspark logger'
backend = ALL_BACKEND  # default value


# public functions
def init():
    """Init Synspark logger."""
    log.info('Init %s...' % NAME)
    _call_backend('init')


def start():
    """Start Synspark logger."""
    log.info('Start %s...' % NAME)
    _call_backend('start')
    api.start()

def stop():
    """Stop Synspark logger."""
    log.info('Stop %s...' % NAME)
    _call_backend('stop')


# private functions
def _call_backend(function_name, *args, **kwargs):
    """Call all function in all backend

    :param function_name: function name
    :param args: function args
    :param kwargs: function kwargs
    """
    for name in list(backend):
        function = getattr(backend[name], function_name, None)
        if function:
            function(*args, **kwargs)


import os
from setuptools import setup, find_packages
from glob import glob

try:
    # python 3
    from distutils.command.build_py import build_py_2to3 as build_py
    from distutils.command.build_scripts \
        import build_scripts_2to3 as build_scripts
except ImportError:
    # python 2
    from distutils.command.build_py import build_py
    from distutils.command.build_scripts import build_scripts

import synspark_logger


# get the long description from the README file
path = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(path, 'README.rst')) as f:
    long_description = f.read()


setup(
    version=synspark_logger.__version__,
    name='synspark_logger',
    packages=find_packages(),
    url='https://www.synspark.com',
    description='Collect your security events to allow you to visualize it on Synspark.',
    long_description=long_description,
    license='',
    platforms='Posix',
    author='Synspark',
    author_email='matthieu.chevrier@synhack.fr',
    data_files=[
        ('etc/synspark_logger', glob('*.cfg')),
        ('usr/share/doc/synspark_logger', ['README.rst']),
    ],
    cmdclass={
        'build_py': build_py,
        'build_scripts': build_scripts,
    },
    scripts=[
        'bin/synspark-logger',
    ],
    install_requires=[
        'requests',
    ],

    # see https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # how mature is this project? Common values are
        #   3 - alpha
        #   4 - beta
        #   5 - production/stable
        'Development Status :: 3 - Alpha',

        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: Log Analysis',
        'Topic :: Security',
        'Topic :: System :: Logging',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ]
)


from synspark_logger.output import log

ALL = {}

try:
    from . import files
    ALL['files'] = files
except ImportError as e:
    log.warning('Disable backend (files): %s' % e)

try:
    from . import systemctl
    ALL['systemctl'] = systemctl
except ImportError as e:
    log.warning('Disable backend (systemctl): %s' % e)



import imp
import os
from subprocess import check_output

from synspark_logger.output import log


files_path = {
    # name: path
    'systemctl': {
        'path': '/lib/systemd/system/synspark_logger.service',
        'post_install_cmd': 'systemctl daemon-reload && '
                            'systemctl start synspark_logger && '
                            'systemctl enable synspark_logger',
        'post_uninstall_cmd': 'systemctl daemon-reload',
        'pre_uninstall_cmd': 'systemctl disable synspark_logger && '
                             'systemctl is-active synspark_logger && '
                             'systemctl stop synspark_logger',
    },
}


def install_daemon():
    log.info('install daemon ...')
    bin_path = check_output(['whereis', 'synspark-logger']).split()[-1].strip()

    template_data = {
        'bin_path': bin_path,
    }

    print(template_data)

    for name, data in files_path.items():
        path = data['path']

        if os.path.exists(path):
            log.warning('%s: "%s" already exist.' % (name, path))
            continue

        _run_command(name, data, 'pre_install_cmd')

        file_name = path.split('/')[-1]
        current_dir_path = os.path.dirname(os.path.abspath(__file__))
        source_path = os.path.join(current_dir_path, 'files', file_name + '.py')

        # for import file with dot
        with open(source_path, 'rb') as f:
            template = imp.load_module('content', f, file_name + '.py',
                ('.py', 'rb', imp.PY_SOURCE)
            ).template

        with open(path, 'w') as f:
            f.write(template % template_data)
            log.info('%s: "%s" created.' % (name, path))

        _run_command(name, data, 'post_install_cmd')

    log.info('install done ...')


def uninstall_daemon():
    log.info('uninstall daemon ...')

    for name, data in files_path.items():
        path = data['path']

        if os.path.exists(path):
            _run_command(name, data, 'pre_uninstall_cmd')

            os.remove(path)
            log.info('%s: "%s" removed.' % (name, path))

        _run_command(name, data, 'post_uninstall_cmd')

    log.info('uninstall done ...')


def _run_command(name, data, cmd_index):
    cmd = data.get(cmd_index)

    if cmd:
        cmd_text = cmd_index.replace('_', ' ')
        log.debug('%s: command: "%s"' % (name, cmd))
        log.info('%s: run %s.' % (name, cmd_text))

        code = os.system(cmd)
        if code != 0:
            log.error('%s: fail %s (code: %d).' % (name, cmd_text, code))


import json
import threading
from datetime import datetime, timedelta
from systemd import journal

from synspark_logger import api
from synspark_logger.output import log

# module vars
__filters = []
__journal = None
__running = False


# public functions
def add_filter(filter):
    """Add filter (rule to allow to send to api)

    :param filter: filter object
    """
    global __filters
    __filters.append(filter)

def init():
    """Init log file watcher."""
    global __journal
    __journal = journal.Reader(converters={'__CURSOR': lambda x: x})
    log.debug('Systemd initialized.')


def start():
    """Start log file watcher."""
    # Seek to now - findtime in journal
    global __running
    __journal.seek_realtime(datetime.now())

    # Move back one entry to ensure do not end up in dead space
    # if start time beyond end of journal
    try:
        __journal.get_previous()
    except OSError:
        pass  # Reading failure, so safe to ignore

    threading.Thread(None, __run, 'backend_systemd').start()
    log.debug('Systemd started.')


def stop():
    """Stop log file watcher."""
    global __journal, __running
    __journal = None
    __running = False
    log.debug('Systemd stopped.')


# private function
class JsonCast(json.JSONEncoder):
    """cast value (json serialize)"""
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        if isinstance(o, timedelta):
            return o.total_seconds()

        return str(o)


def __run():
    """Running function"""
    global __running
    __running = True
    while __running:
        try:
            log_entry = __journal.get_next()
        except OSError:
            log.warning('Systemd: error on reading line from journal')
            continue

        # wait event (timeout on 1s for check if thread running)
        if not log_entry:
            __journal.wait(1)
            continue

        # is allowned to send to api ?
        allowed = False
        for filter in __filters:
            if filter.check(
                message=log_entry['MESSAGE'],
                criticity=log_entry['PRIORITY'],
                syslog_facility=log_entry['SYSLOG_FACILITY'],
            ):
                allowed = True
                break

        else:  # no filter => all is allowed
            allowed = True

        if allowed:
            api.send(log=json.dumps(log_entry, cls=JsonCast))
